import json
import logging
import os
import shutil
import subprocess
import sys
from multiprocessing import Process
from pathlib import Path

import requests

from FaaSr_py.config.debug_config import global_config
from FaaSr_py.engine.faasr_payload import FaaSrPayload
from FaaSr_py.helpers.faasr_start_invoke_helper import faasr_func_dependancy_install
from FaaSr_py.helpers.s3_helper_functions import flush_s3_log, get_invocation_folder
from FaaSr_py.s3_api import faasr_put_file
from FaaSr_py.server.faasr_server import run_server, wait_for_server_start

logger = logging.getLogger(__name__)


class Executor:
    """
    Handles logic related to running user function
    """

    def __init__(self, faasr: FaaSrPayload):
        if not isinstance(faasr, FaaSrPayload):
            err_msg = "initializer for Executor must be FaaSr instance"
            logger.error(err_msg)
            sys.exit(1)
        self.faasr = faasr
        self.server = None
        self.packages = []

    def _call(self, action_name):
        """
        Runs a user function given action name

        Arguments:
            action_name: str -- name of the action to run
        """
        func_name = self.faasr["ActionList"][action_name]["FunctionName"]
        func_type = self.faasr["ActionList"][action_name]["Type"]
        user_args = self._get_user_function_args(action_name)

        if not global_config.SKIP_USER_FUNCTION:
            if func_type == "Python":
                # entry script for py function
                from FaaSr_py.client.py_user_func_entry import run_py_function

                # run user func as seperate process
                try:
                    py_func = Process(
                        target=run_py_function,
                        args=(self.faasr, func_name, user_args),
                    )
                except Exception as e:
                    logger.error(f"Error running Python function: {e}")
                    sys.exit(1)

                logger.info(f"Starting function: {func_name} (Python)")
                py_func.start()
                py_func.join()

                func_res = py_func.exitcode
            elif func_type == "R":
                # path to R function handler
                client_dir = Path(__file__).parent.parent / "client"

                r_files = [
                    client_dir / "r_user_func_entry.R",
                    client_dir / "r_func_helper.R",
                    client_dir / "r_client_stubs.R",
                ]

                # Copy each file into the current working directory
                for src in r_files:
                    dst = Path(src.name)
                    shutil.copy(src, dst)

                logger.info(f"Starting function: {func_name} (R)")

                # run R entry as a subprocess
                try:
                    r_func = subprocess.run(
                        [
                            "Rscript",
                            "r_user_func_entry.R",
                            func_name,
                            json.dumps(user_args),
                            self.faasr["InvocationID"],
                        ]
                    )
                except Exception as e:
                    logger.error(f"Error running R function: {e}")
                    sys.exit(1)
                func_res = r_func.returncode
            else:
                logger.error(f"Unkown function type: {func_type}")
                sys.exit(1)

            if func_res != 0:
                raise RuntimeError(
                    f"non-zero exit code ({func_res!r}) from user function"
                )
        else:
            logger.info("SKIPPING USER FUNCTION")

        self._make_done(action_name)

    def _make_done(self, action_name):
        """
        At this point, the action has finished the invocation of the user Function
        We flag this by uploading a file with the name {action}.done to the S3 logs folder
        Check if directory already exists. If not, create one
        """
        log_folder = get_invocation_folder(self.faasr)
        log_folder_path = f"/tmp/{log_folder}/{action_name}/flag"

        if not os.path.isdir(f"{log_folder_path}/function_completions"):
            os.makedirs(f"{log_folder_path}/function_completions")

        if "FunctionRank" in self.faasr:
            file_name = (
                f"function_completions/{action_name}.{self.faasr['FunctionRank']}.done"
            )
        else:
            file_name = f"function_completions/{action_name}.done"

        # Set content of .done file to True
        with open(f"{log_folder_path}/{file_name}", "w") as f:
            f.write("True")

        # Put .done file in S3
        faasr_put_file(
            faasr_payload=self.faasr,
            local_folder=log_folder_path,
            local_file=file_name,
            remote_folder=log_folder,
            remote_file=file_name,
        )

        logger.debug(f"Put {file_name} file in S3")

    def _run_builtin_function(self, action_name, action_config):
        """
        Execute a FaaSr built-in function.
        Built-ins don't need dependency installation, server, or subprocess execution.
        
        Args:
            action_name: Name of the action
            action_config: Action configuration
            
        Returns:
            Function result
        """
        builtin_func_name = action_config.get("FunctionName")
        
        if not builtin_func_name:
            raise ValueError(f"Built-in action {action_name} missing FunctionName")
        
        try:
            # Import built-in functions
            from FaaSr_py.builtin_functions import vm_start, vm_stop
            
            builtin_functions = {
                "vm_start": vm_start,
                "vm_stop": vm_stop
            }
            
            if builtin_func_name not in builtin_functions:
                raise ValueError(f"Unknown built-in function: {builtin_func_name}")
            
            func = builtin_functions[builtin_func_name]
            
            logger.info(f"Running built-in: {builtin_func_name}")
            result = func(self.faasr)
            
            # Create .done file for synchronization (same as user functions)
            self._make_done(action_name)
            
            logger.info(f"Built-in function {builtin_func_name} completed successfully")
            return result
            
        except Exception as e:
            logger.error(f"Built-in function {builtin_func_name} failed: {e}")
            raise

    def run_func(self, action_name, start_time):
        """
        Fetch and run the users function

        Arguments:
            action_name: str -- name of the action to run
        """

        action_config = self.faasr["ActionList"].get(action_name, {})
        if action_config.get("_faasr_builtin", False):
            logger.info(f"Executing built-in function: {action_name}")
            return self._run_builtin_function(action_name, action_config)
        
        # For user functions, continue with existing logic
        logger.debug("Starting dependency install")
        action = self.faasr["ActionList"][action_name]
        faasr_func_dependancy_install(self.faasr, action)
        # install dependencies for function
        logger.debug("Starting dependency install")
        action = self.faasr["ActionList"][action_name]
        faasr_func_dependancy_install(self.faasr, action)
        logger.debug("Finished installing dependencies")

        # Run function
        try:
            self._host_server_api(start_time=start_time)
            self._call(action_name)
            function_result = self.get_function_return()
        except Exception as e:
            if isinstance(e, SystemExit):
                raise
            logger.exception(e, stack_info=True)
            sys.exit(1)
        finally:
            # Clean up server
            self.terminate_server()
        return function_result

    def _host_server_api(self, start_time, port=8000):
        """
        Starts RPC server for serverside API

        Arguments:
            port: int -- port to run the server on
        """
        logger.info(f"Starting server on localhost port {port}")
        # flush s3 log since server process will be logging
        flush_s3_log()
        self.server = Process(target=run_server, args=(self.faasr, port, start_time))
        self.server.start()
        logger.debug("Polling localhost")
        wait_for_server_start(port)

    def terminate_server(self):
        """
        Terminate RPC server
        """
        if isinstance(self.server, Process):
            self.server.terminate()
        else:
            err_msg = "Tried to terminate server, but no server running"
            logger.error(err_msg)
            sys.exit(1)

    def _get_user_function_args(self, action_name):
        """
        Returns user function arguments

        Returns:
            dict -- user function arguments
        """
        args = self.faasr["ActionList"][action_name]["Arguments"]
        if args is None:
            return {}
        else:
            return args

    def get_function_return(self, port=8000):
        """
        Get user function result

        Arguments:
            port: int -- port to get the function result from

        Returns:
            result: bool | None
        """
        try:
            return_response = requests.get(f"http://127.0.0.1:{port}/faasr-get-return")
            return_val = return_response.json()
        except Exception:
            err_msg = "Error getting function result"
            logger.exception(err_msg, stack_info=True)
            sys.exit(1)

        if return_val.get("Error"):
            if return_val.get("Message"):
                err_msg = f"{return_val['Message']}"
            else:
                err_msg = "Unkown error while getting user function return"
            logger.error(err_msg, stack_info=True)
            raise RuntimeError(err_msg)

        return return_val["FunctionResult"]



from pathlib import Path

from FaaSr_py.config.debug_config import global_config


def default_func():
    # Create test file
    test_file = Path("/tmp") / "test_upload.txt"
    test_file.write_text("This is a test file for FaaSr.\nLine 2 of the file.")

    # Upload file using faasr_put_file
    faasr_put_file(  # noqa: F821
        local_file="test_upload.txt",
        remote_file="uploaded.txt",
        local_folder="/tmp",
        remote_folder="test-folder",
        server_name="local",
    )

    # Print uploaded file content from the local "bucket"
    uploaded_path = (
        Path(global_config.LOCAL_FILE_SYSTEM_DIR) / "test-folder" / "uploaded.txt"
    )
    if uploaded_path.exists():
        print("\nUploaded file contents:")
        print("-" * 30)
        print(uploaded_path.read_text())
        print("-" * 30)
    else:
        print("Uploaded file not found.")

    # Delete file using faasr_delete_file
    faasr_delete_file(  # noqa: F821
        remote_file="uploaded.txt", remote_folder="test-folder", server_name="local"
    )

    # Confirm deletion
    if not uploaded_path.exists():
        print("File successfully deleted.")
    else:
        print("File still exists after deletion.")



import json
import logging
import sys
from collections import defaultdict
from pathlib import Path

from jsonschema import validate
from jsonschema.exceptions import ValidationError

logger = logging.getLogger(__name__)


def validate_json(payload):
    """
    Verifies JSON payload is compliant with the FaaSr schema

    Arguments:
        payload: FaaSr payload to validate
    """
    if isinstance(payload, str):
        payload = json.loads(payload)

    schema_path = Path(__file__).parent.parent / "FaaSr.schema.json"
    if not schema_path.exists():
        logger.error(f"FaaSr schema file not found at {schema_path}")
        sys.exit(1)

    # Open FaaSr schema
    with open(schema_path, "r") as f:
        schema = json.load(f)

    # Compare payload against FaaSr schema and except if they do not match
    try:
        validate(instance=payload, schema=schema)
    except ValidationError as e:
        logger.error(f"JSON not compliant with FaaSr schema: {e.message}")
        sys.exit(1)
    return True


def is_cyclic(adj_graph, curr, visited, stack):
    """
    Recursive function that if there is a cycle in a directed
    graph defined by an adjacency list

    Arguments:
        adj_graph: adjacency list for graph (dict)
        curr: current node
        visited: set of visited nodes (set)
        stack: list of nodes in recursion call stack (list)

    Returns:
        bool: True if cycle exists, False otherwise
    """
    # if the current node is in the recursion call
    # stack then there must be a cycle in the graph
    if curr in stack:
        return True

    # add current node to recursion call stack and visited set
    visited.add(curr)
    stack.append(curr)

    # check each successor for cycles, recursively calling is_cyclic()
    for child in adj_graph[curr]:
        if child not in visited and is_cyclic(adj_graph, child, visited, stack):
            logger.error(f"Function loop found from node {curr} to {child}")
            sys.exit(1)
        elif child in stack:
            logger.error(f"Function loop found from node {curr} to {child}")
            sys.exit(1)

    # no more successors to visit for this branch and no cycles found
    # remove current node from recursion call stack
    stack.pop()
    return False


def build_adjacency_graph(payload):
    """
    This function builds an adjacency list for the FaaSr workflow graph and determines
    the ranks of each action

    Arguments:
        payload: FaaSr payload dict
    Returns:
        adj_graph: dict of predecessor: succesor pairs
        rank: dict of each action's rank
    """
    adj_graph = defaultdict(list)

    ranks = dict()

    # Build adjacency list from ActionList
    for func in payload["ActionList"].keys():
        invoke_next = payload["ActionList"][func]["InvokeNext"]
        if isinstance(invoke_next, str):
            invoke_next = [invoke_next]
        for child in invoke_next:

            def process_action(action):
                action_name, action_rank = extract_rank(action)
                if action_name in ranks and ranks[action_name] > 1:
                    err_msg = "Function with rank cannot have multiple predecessors"
                    logger.error(err_msg)
                    sys.exit(1)
                else:
                    adj_graph[func].append(action_name)
                    ranks[action_name] = action_rank

            if isinstance(child, dict):
                for conditional_branch in child.values():
                    for action in conditional_branch:
                        process_action(action)
            else:
                process_action(child)

    for func in adj_graph:
        if func not in ranks:
            ranks[func] = 0

    return (adj_graph, ranks)


def get_ranks(payload):
    """Returns just dict mapping functions to their rank"""
    _, rank = build_adjacency_graph(payload)
    return rank


def check_dag(faasr_payload):
    """
    This method checks for cycles, repeated function names,
    or unreachable nodes in the workflow and aborts if it finds any

    Arguments:
        payload: FaaSr payload dict
    Returns:
        predecessors: dict -- map of function predecessors
    """
    if faasr_payload["FunctionInvoke"] not in faasr_payload["ActionList"]:
        err_msg = "FunctionInvoke does not refer to a valid function"
        logger.error(err_msg)
        sys.exit(1)

    adj_graph, ranks = build_adjacency_graph(faasr_payload)

    # Initialize empty recursion call stack
    stack = []

    # Initialize empty visited set
    visited = set()

    # Find initial function in the graph
    start = False
    for func in faasr_payload["ActionList"]:
        if ranks[func] == 0:
            start = True
            # This function stores the first function with no predecessors
            # In the cases where there is multiple functions with no
            # predecessors, an unreachable state error will occur later
            first_func = func
            break

    # Ensure there is an initial action
    if start is False:
        logger.error("Function loop found: no initial action")
        sys.exit(1)

    # Check for cycles
    is_cyclic(adj_graph, first_func, visited, stack)

    # Check if all of the functions have been visited by the DFS
    # If not, then there is an unreachable state in the graph
    for func in faasr_payload["ActionList"]:
        if func.split(".")[0] not in visited:
            logger.error(f"Unreachable state found: {func}")
            sys.exit(1)

    # Initialize predecessor list
    pre = predecessors_list(adj_graph)

    # Ensure that no ranked function invokes another ranked function
    for func, p in pre.items():
        if ranks[func] > 1:
            for pre_f in p:
                if ranks[pre_f] > 1:
                    logger.error(
                        "Function with rank cannot have predecessor with rank"
                        f" - offending functions: {func}({ranks[func]}) and {pre_f}({ranks[pre_f]})"
                    )
                    sys.exit(1)

    curr_pre = pre[faasr_payload["FunctionInvoke"]]
    real_pre = []
    for p in curr_pre:
        if p in ranks and ranks[p] > 1:
            for i in range(1, ranks[p] + 1):
                real_pre.append(f"{p}.{i}")
        else:
            real_pre.append(p)
    return real_pre


def predecessors_list(adj_graph):
    """This function returns a map of action predecessor pairs

    Arguments:
        adj_graph: adjacency list for graph -- dict(function: successor)
    """
    pre = defaultdict(list)
    for func1 in adj_graph:
        for func2 in adj_graph[func1]:
            pre[func2].append(func1)
    return pre


def extract_rank(str):
    """
    Returns action name and rank of an action with rank (e.g func(7) returns (func, 7))

    Arguments:
        str: function name with rank
    Returns:
        (str, int) -- action name and rank
    """
    parts = str.split("(")
    if len(parts) != 2 or not parts[1].endswith(")"):
        return str, 1
    rank = int(parts[1][:-1])
    action_name = parts[0]
    return (action_name, rank)



from setuptools import find_packages, setup

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="FaaSr_py",
    version="0.1.13",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
)



import logging

from FaaSr_py.config.s3_log_sender import S3LogSender

logger = logging.getLogger(__name__)


class S3LogHandler(logging.Handler):
    """
    Handler for Python logger to send logs to S3
    """

    def __init__(self, faasr_payload, level, start_time):
        """
        Arguments:
            faasr_payload: FaaSr payload instance
            level: minimum logging threshold (DEBUG, INFO, WARNING, ERROR, EXCEPTION)
            start_time: timestamp from start of FaaSr action
        """
        # Initialize the S3LogSender with the provided faasr_payload
        self._sender = S3LogSender(timestamp=start_time, faasr_payload=faasr_payload)
        super().__init__(level=level)

    def emit(self, record):
        try:
            # get timestamp since start of func
            record.timestamp = self._sender.get_curr_timestamp()
            if self.formatter is None:
                formatter = logging.Formatter(
                    "[%(timestamp)s] [%(levelname)s] [%(filename)s] %(message)s"
                )
                self.setFormatter(formatter)

            # format log and send it
            msg = self.format(record)
            self._sender.log(msg)
        except Exception as e:
            self._sender.flush_log()
            raise RuntimeError("failed to upload s3 log") from e

        # flush log if it is an error
        if record.levelno >= logging.ERROR:
            self._sender.flush_log()



import logging
import sys
from pathlib import Path

import boto3

from FaaSr_py.config.debug_config import global_config

logger = logging.getLogger(__name__)


def faasr_get_folder_list(faasr_payload, server_name="", prefix=""):
    """
    Get a list of objects in the S3 bucket

    Arguments:
        faasr_payload: FaaSr payload dict
        server_name: str -- name of S3 data store to get folder list from
        prefix: str -- prefix to filter objects in S3 bucket
    Returns:
        list: List of objects in the S3 bucket with the specified prefix
    """

    if global_config.USE_LOCAL_FILE_SYSTEM:
        logger.info("Getting folder list from local bucket")

        local_bucket = Path(global_config.LOCAL_FILE_SYSTEM_DIR)
        folder_path = local_bucket / prefix

        all_files = [p for p in folder_path.rglob("*") if p.is_file()]

        stripped_files = [str(p.relative_to(local_bucket.parent)) for p in all_files]
        print(stripped_files)
    else:
        # Get server name from payload if one is not providedS
        if server_name == "":
            server_name = faasr_payload["DefaultDataStore"]

        # Ensure the server is a valid data store
        if server_name not in faasr_payload["DataStores"]:
            logger.error("Invalid data server name: {server_name}")
            sys.exit(1)

        # Get the S3 data store to get folder list from
        target_s3 = faasr_payload["DataStores"][server_name]

        if target_s3.get("Endpoint"):
            s3_client = boto3.client(
                "s3",
                aws_access_key_id=target_s3["AccessKey"],
                aws_secret_access_key=target_s3["SecretKey"],
                region_name=target_s3["Region"],
                endpoint_url=target_s3["Endpoint"],
            )
        else:
            s3_client = boto3.client(
                "s3",
                aws_access_key_id=target_s3["AccessKey"],
                aws_secret_access_key=target_s3["SecretKey"],
                region_name=target_s3["Region"],
            )

        # List objects from S3 bucket
        result = s3_client.list_objects_v2(
            Bucket=target_s3["Bucket"], Prefix=str(prefix)
        )
        if "Contents" in result:
            result = [content["Key"] for content in result["Contents"]]
            result = [obj for obj in result if not obj.endswith("/")]
            return result
        return []



"""
Utilities for determining function position in workflow DAG.
"""

import logging

logger = logging.getLogger("FaaSr_py.vm")

def does_next_action_require_vm(faasr_payload, current_action):
    """
    Check if any immediate next action requires VM.
    
    Args:
        faasr_payload: The FaaSr workflow configuration
        current_action: Name of current action
        
    Returns:
        bool: True if any next action requires VM
    """
    action_list = faasr_payload.get("ActionList", {})
    current_config = action_list.get(current_action, {})
    next_actions = current_config.get("InvokeNext", [])
    
    # Handle single string or list
    if isinstance(next_actions, str):
        next_actions = [next_actions]
    elif not isinstance(next_actions, list):
        next_actions = []
    
    # Check each next action
    for next_action in next_actions:
        if action_list.get(next_action, {}).get("RequiresVM", False):
            return True
    
    return False



def get_action_position(faasr_payload, action_name):
    """
    Determine if an action is first, last, or middle in the workflow.
    
    Args:
        faasr_payload: The FaaSr workflow configuration payload
        action_name: Name of the action to check
        
    Returns:
        dict: Position information with is_first, is_last, and type fields
    """
    action_list = faasr_payload.get("ActionList", {})
    
    # Find all actions that invoke this action
    invoking_actions = []
    for name, config in action_list.items():
        invoke_next = config.get("InvokeNext", [])
        if isinstance(invoke_next, list) and action_name in invoke_next:
            invoking_actions.append(name)
        elif isinstance(invoke_next, dict):
            # Handle conditional invocations
            for condition in ["True", "False"]:
                if condition in invoke_next and action_name in invoke_next[condition]:
                    invoking_actions.append(name)
    
    # Find actions this action invokes
    next_actions = action_list.get(action_name, {}).get("InvokeNext", [])
    
    # Determine position
    is_first = len(invoking_actions) == 0
    is_last = not next_actions
    
    position_type = "middle"
    if is_first and is_last:
        position_type = "only"
    elif is_first:
        position_type = "first"
    elif is_last:
        position_type = "last"
    
    return {
        "is_first": is_first,
        "is_last": is_last,
        "type": position_type
    }

def get_vm_requiring_actions(faasr_payload):
    """
    Get a list of all actions that require VM.
    
    Args:
        faasr_payload: The FaaSr workflow configuration payload
        
    Returns:
        list: Names of actions requiring VM
    """
    vm_actions = []
    
    for name, config in faasr_payload.get("ActionList", {}).items():
        if config.get("RequiresVM", False):
            vm_actions.append(name)
    
    return vm_actions

def is_first_vm_action(faasr_payload, action_name):
    """
    Check if action is the first VM-requiring action in execution flow.
    
    Args:
        faasr_payload: The FaaSr workflow configuration payload
        action_name: Name of the action to check
        
    Returns:
        bool: True if it's the first VM action, False otherwise
    """
    from .detection import action_requires_vm
    
    if not action_requires_vm(faasr_payload, action_name):
        return False
    
    position = get_action_position(faasr_payload, action_name)
    
    if position["is_first"]:
        return True
        
    # If not the first action in workflow, check if it's the first VM-requiring action
    for name, config in faasr_payload.get("ActionList", {}).items():
        # Skip the current action
        if name == action_name:
            continue
            
        # If we find another VM-requiring action that runs before this one, this isn't first
        if config.get("RequiresVM", False):
            position = get_action_position(faasr_payload, name)
            if position["is_first"]:
                return False
    
    return True

def is_last_vm_action(faasr_payload, action_name):
    """
    Check if action is the last VM-requiring action in execution flow.
    
    Args:
        faasr_payload: The FaaSr workflow configuration payload
        action_name: Name of the action to check
        
    Returns:
        bool: True if it's the last VM action, False otherwise
    """
    from .detection import action_requires_vm
    
    if not action_requires_vm(faasr_payload, action_name):
        return False
    
    position = get_action_position(faasr_payload, action_name)
    
    if position["is_last"]:
        return True
        
    # If not the last action in workflow, check if it's the last VM-requiring action
    for name, config in faasr_payload.get("ActionList", {}).items():
        # Skip the current action
        if name == action_name:
            continue
            
        # If we find another VM-requiring action that runs after this one, this isn't last
        if config.get("RequiresVM", False):
            position = get_action_position(faasr_payload, name)
            if position["is_last"]:
                return False
    
    return True


import logging
from pathlib import Path

from FaaSr_py.client.py_client_stubs import (
    faasr_delete_file,
    faasr_exit,
    faasr_get_file,
    faasr_get_folder_list,
    faasr_get_s3_creds,
    faasr_log,
    faasr_put_file,
    faasr_rank,
    faasr_return,
)
from FaaSr_py.config.debug_config import global_config
from FaaSr_py.helpers.py_func_helper import (
    faasr_import_function,
    faasr_import_function_walk,
    local_wrap,
)

logger = logging.getLogger(__name__)


def run_py_function(faasr, func_name, args, func_path=None):
    """
    Entry for Python function process

    Arguments:
        faasr: FaaSr payload instance
        func_name: name of function to run
        args: arguments for function (dict)
    """
    try:
        if global_config.USE_LOCAL_USER_FUNC:
            func_path = Path(global_config.LOCAL_FUNCTION_PATH).resolve()
            func_name = global_config.LOCAL_FUNCTION_NAME

            user_function = faasr_import_function(func_path, func_name)
        else:
            user_function = faasr_import_function_walk(
                func_name, directory=f"/tmp/functions/{faasr['InvocationID']}"
            )
    except Exception as e:
        err_msg = f"failed to get user functions -- error: {e}"
        faasr_exit(err_msg)

    # Ensure user function is present
    if not user_function:
        err_msg = f"{{py_user_func_entry.py: cannot find function {func_name}}}"
        faasr_exit(err_msg)

    # Add FaaSr client stubs to user function's namespace
    user_function.__globals__["faasr_put_file"] = faasr_put_file
    user_function.__globals__["faasr_get_file"] = faasr_get_file
    user_function.__globals__["faasr_delete_file"] = faasr_delete_file
    user_function.__globals__["faasr_get_folder_list"] = faasr_get_folder_list
    user_function.__globals__["faasr_log"] = faasr_log
    user_function.__globals__["faasr_rank"] = faasr_rank
    user_function.__globals__["faasr_get_s3_creds"] = faasr_get_s3_creds

    try:
        if global_config.USE_LOCAL_USER_FUNC:
            print(f"using local function {global_config.LOCAL_FUNCTION_NAME}")
            result = local_wrap(user_function)(**global_config.LOCAL_FUNC_ARGS)
        else:
            result = user_function(**args)
    except Exception as e:
        faasr_exit(message=str(e))

    faasr_return(result)
